from django import template

register = template.Library()

@register.filter(name='calculate_discount')
def calculate_discount(price, percent=None):
    try:
        price = round(float(price), 2)
        percent = int(percent)
    except:
        return price
    
    discounted_price = price - (price*(percent/100))
    return round(discounted_price, 2)



from django.urls import path
from . import views

app_name = 'store'

urlpatterns = [
    path('', views.homepage, name='homepage'),
    path('search/', views.search, name='search'),
    path('apply-discount/', views.apply_discount),
    path('add-to-cart/', views.add2cart),
    path('remove-from-cart/', views.removeFromCart),
    path('get-cart/', views.getCart),
    path('clear-cart/', views.clearCart),
    path('get-products/', views.getProducts),
    path('check-deliverability/', views.checkDeliverability),

    path('checkout/', views.checkout, name="checkout"),
    path('checkout/select-delivery-address/', views.selectDeliveryAddress),
    path('checkout/remove-selected-delivery-address/', views.removeSelectedDeliveryAddress),
    path('checkout/get-delivery-address/', views.getDeliveryAddress),
    path('checkout/place-order/', views.placeOrder),

    path('orders/', views.orders, name="orders"),
    path('orders/cancel/', views.cancelOrder),
    path('orders/<int:order_id>/', views.order, name="order"),
    
    path('profile/', views.profile, name='profile'),
    path('profile/get-user/', views.getUser),
    path('profile/add-address/', views.addAddress),
    path('profile/make-address-primary/', views.makeAddressPrimary),
    path('profile/delete-address/', views.deleteAddress),
    path('profile/get-address/', views.getAddress),
    path('profile/edit-address/', views.editAddress),

    path('<str:product_slug>/', views.products, name='products'),
    path('<str:product_slug>/<int:variation_id>/', views.product, name='product')
]


from django.contrib import admin
from . import models
from django.db.models import Count
from django.utils.html import format_html, urlencode
from django.urls import reverse
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.contrib.auth.models import User
from datetime import datetime, timezone


class UserAdmin(BaseUserAdmin):
    list_display = ["username", "email", "first_name", "last_name", "is_staff", "orders_count"]

    @admin.display(ordering="orders_count")
    def orders_count(self, user):
        url = reverse("admin:store_order_changelist") + '?' + urlencode({
            "customer__id": str(user.id)
        })
        return format_html("<a href='{}'>{}</a>", url, user.orders_count) 
    
    def get_queryset(self, request):
        return super().get_queryset(request).annotate(
            orders_count=Count("orders")
        )


admin.site.unregister(User)
admin.site.register(User, UserAdmin)


### Orders
@admin.register(models.Order)
class OrderAdmin(admin.ModelAdmin):
    actions = ["mark_closed", "cancel_order", "process_cancellation"]
    list_display = ["id", "customer", "amount_payable", "order_status", "items_count", "placed_at", "expected_delivery_time"]
    list_select_related = ["discount", "cancellation", "customer"]
    search_fields = ["id"]
    list_filter = ["placed_at", "order_status"]
    list_per_page = 20


    @admin.display(ordering="items_count")
    def items_count(self, order):
        url = reverse("admin:store_orderitem_changelist") + '?' + urlencode({
            "order__id": str(order.id)
        })
        return format_html("<a href='{}'>{}</a>", url, order.items_count) 
    
    def get_queryset(self, request):
        return super().get_queryset(request).annotate(
            items_count=Count("orderitems")
        )
    
    @admin.action(description="Mark as Closed")
    def mark_closed(self, request, queryset):
        updated_count = 0
        for order in queryset:
            if order.order_status == 'O':
                order.order_status = 'C'
                order.payment_status = 'S'
                order.closed_at = datetime.now(timezone.utc)
                order.save()
                updated_count += 1
        self.message_user(
            request,
            f"{updated_count} orders were successfully closed."
        )
    

    @admin.action(description="Cancel Order")
    def cancel_order(self, request, queryset):
        updated_count = 0
        for order in queryset:
            if order.order_status == 'P':
                order.order_status = 'N'
                order.payment_status = 'U'
                order.cancelled_at = datetime.now(timezone.utc)
                order.save()
                updated_count += 1

                for item in order.orderitems.select_related("product").all():
                    item.product.inventory += item.quantity
                    if item.product.inventory > 0:
                        item.product.availability = True
                    item.product.save()
        self.message_user(
            request,
            f"{updated_count} orders were successfully cancelled."
        )


    @admin.action(description="Process Cancellation")
    def process_cancellation(self, request, queryset):
        updated_count = 0
        for order in queryset:
            if order.order_status == 'O':
                order.order_status = 'P'
                order.payment_status = 'U'
                order.save()
            
                oc = models.OrderCancellation.objects.create(
                    order = order
                )
                updated_count += 1
        self.message_user(
            request,
            f"Processing cancellation for {updated_count} orders."
        )



### Order Items
@admin.register(models.OrderItem)
class OrderItemAdmin(admin.ModelAdmin):
    list_display = ["id", "order", "order_customer", "product", "quantity", "amount_payable"]
    list_select_related = ["order", "order__customer", "product", "product__product"]
    readonly_fields = ["order", "product", "discount_percent"]
    search_fields = ["order__id"]
    list_per_page = 20
    
    def order_customer(self, orderitem):
        return orderitem.order.customer


### Cancelled Orders
@admin.register(models.OrderCancellation)
class OrderCancellationAdmin(admin.ModelAdmin):
    list_display = ["order", "cancelled_at", "order_customer"]
    readonly_fields = ["order"]
    search_fields = ["order__id"]
    list_per_page = 20
    
    def order_customer(self, cancelleditem):
        return cancelleditem.order.customer


### Address
@admin.register(models.Address)
class AddressAdmin(admin.ModelAdmin):
    list_filter = ["user", "city", "state", "country"]
    list_display = ["user", "address1", "city", "pincode", "mobile"]
    empty_value_display = '-empty-'
    search_fields = ["address1", "address2", "landmark"]
    list_per_page = 20


########## Image ##########
@admin.register(models.Image)
class ImageAdmin(admin.ModelAdmin):
    list_display = ['image', 'name']
    list_editable = ['name']
    list_per_page = 10
    search_fields = ['name']


### Product Tags
@admin.register(models.Tag)
class TagAdmin(admin.ModelAdmin):
    list_display = ['id', 'name']
    list_editable = ['name']
    ordering = ['name']
    search_fields = ['name']
    list_per_page = 20


### Category
@admin.register(models.Category)
class CategoryAdmin(admin.ModelAdmin):
    list_display = ['id', 'title', 'product_count']
    list_per_page = 10
    ordering = ['title']
    search_fields = ['title']
    prepopulated_fields = {
        'slug': ['title']
    }

    @admin.display(ordering='product_count')
    def product_count(self, category):
        url = (
            reverse('admin:store_product_changelist')
            + '?'
            + urlencode({
                'categories__id': str(category.id)
            })
        )
        return format_html(
            '<a href="{}">{}</a>',
            url,
            category.product_count
        )
    
    def get_queryset(self, request):
        return super().get_queryset(request).annotate(
            product_count=Count('products')
        )


### Pincode
@admin.register(models.Pincode)
class PincodeAdmin(admin.ModelAdmin):
    list_display = ["pincode", "delivery_time_in_days"]
    list_editable = ["delivery_time_in_days"]
    autocomplete_fields = ["products_unavailable"]
    list_per_page = 10
    search_fields = ["pincode__startswith"]


### Product
@admin.register(models.Product)
class ProductAdmin(admin.ModelAdmin):
    list_display = ['title', 'variation_count']
    list_per_page = 10
    list_filter = ['categories__title']
    search_fields = ['title__istartswith']
    filter_horizontal = ["tags", "categories"]
    prepopulated_fields = {
        'slug': ['title']
    }

    @admin.display(ordering='variation_count')
    def variation_count(self, product):
        url = (
            reverse('admin:store_variation_changelist')
            + '?'
            + urlencode({
                'product__id': str(product.id)
            })
        )
        return format_html(
            '<a href="{}">{}</a>',
            url,
            product.variation_count
        )
    
    def get_queryset(self, request):
        return super().get_queryset(request).prefetch_related(
            "categories", "tags"
        ).annotate(
            variation_count=Count('variations')
        )


### Inventory Filter for Variations
class InventoryFilter(admin.SimpleListFilter):
    title = 'inventory'
    parameter_name = 'inventory'

    def lookups(self, request, model_admin):
        return [
            ('<10', "Low")
        ]
    
    def queryset(self, request, queryset):
        if self.value() == '<10':
            return queryset.filter(inventory__lt=10)


### Variation
@admin.register(models.Variation)
class VariationAdmin(admin.ModelAdmin):
    autocomplete_fields = ['product']
    list_display = ['product', 'price', 'unit', 'availability', 'inventory', 'inventory_status', 'orders_count']
    list_editable = ['price', 'availability', 'inventory']
    list_per_page = 10
    list_filter = [InventoryFilter, 'keyvalues__value']
    search_fields = ['product__title__istartswith']
    filter_horizontal = ["images", "keyvalues"]
    list_select_related = [models.Product, models.Variation.keyvalues, models.Variation.images, models.KeyValue.key, models.KeyValue.value]
    
    @admin.display(ordering='inventory')
    def inventory_status(self, variation):
        if variation.inventory < 10:
            return 'Low'
        return 'OK'
    
    @admin.display(ordering="orders_count")
    def orders_count(self, variation):
        return variation.orders_count
    
    def get_queryset(self, request):
        return super().get_queryset(request).select_related("product").annotate(
            orders_count=Count("orderitems")
        )


### Discount
@admin.register(models.Discount)
class DiscountAdmin(admin.ModelAdmin):
    actions = ['clear_discounts']
    list_display = ['id', 'code', 'percent']
    list_editable = ['code', 'percent']
    search_fields = ["percent"]
    list_per_page = 20

    @admin.action(description='Clear Discounts')
    def clear_discounts(self, request, queryset):
        updated_count = queryset.update(percent=0)
        self.message_user(
            request,
            f'{updated_count} discounts were successfully updated.'
        )


### Keys
@admin.register(models.Key)
class KeyAdmin(admin.ModelAdmin):
    list_display = ['key', 'value_count']
    search_fields = ['key']

    @admin.display(ordering='value_count')
    def value_count(self, key):
        url = (
            reverse('admin:store_value_changelist')
            + '?'
            + urlencode({
                'key__id': str(key.id)
            })
        )
        return format_html(
            '<a href="{}">{}</a>',
            url,
            key.value_count
        )
    
    def get_queryset(self, request):
        return super().get_queryset(request).annotate(
            value_count=Count('values')
        )


### Value
@admin.register(models.Value)
class ValueAdmin(admin.ModelAdmin):
    search_fields = ['value']


### Key - Values
@admin.register(models.KeyValue)
class KeyValueAdmin(admin.ModelAdmin):
    autocomplete_fields = ['key', 'value']
    list_display = ['id', 'key_title', 'value_title', 'product_count']
    list_select_related = ['key', 'value']
    search_fields = ['key__key__istartswith', 'value__value__istartswith']

    @admin.display(ordering='key')
    def key_title(self, keyvalue):
        return keyvalue.key.key
    
    @admin.display(ordering='value')
    def value_title(self, keyvalue):
        return keyvalue.value.value
    
    @admin.display(ordering='product_count')
    def product_count(self, keyvalue):
        url = (
            reverse('admin:store_variation_changelist')
            + '?'
            + urlencode({
                'keyvalues__id': str(keyvalue.id)
            })
        )
        return format_html(
            '<a href="{}">{}</a>',
            url,
            keyvalue.product_count
        )
    
    def get_queryset(self, request):
        return super().get_queryset(request).annotate(
            product_count=Count('variation')
        )



from .models import Order, Product, Category, Variation, Pincode, Tag
from datetime import datetime
from django.core.paginator import Paginator
from django.urls import reverse
from django.conf import settings
from django.db.models import Count, F, Value

def get_search_products(request):
    q = request.GET.get('q')
    if not q:
        return None
    products = Product.objects.filter(title__istartswith=q).values("id")
    tags = Tag.objects.prefetch_related("products").filter(name__istartswith=q)
    for tag in tags:
        products = products.union(tag.products.prefetch_related("variations").all())
    
    if not products:
        return {"products": None, "search_query": q}
    
    variations = Variation.objects.select_related("product").filter(product__in=products)
    variations = variations.select_related("discount").prefetch_related(
        "keyvalues",
        "images",
        "keyvalues__key",
        "keyvalues__value"
    ).filter(availability=True)

    selected_sort = None
    if request.GET.get('sort'):
        sort = request.GET.get('sort')
        if sort == "z-a":
            variations = variations.order_by("-product")
            selected_sort = {"value": sort, "display": "Z - A"}
        elif sort == "price-asc":
            variations = variations.order_by("price", "product")
            selected_sort = {"value": sort, "display": "Price (Low to High)"}
        elif sort == "price-desc":
            variations = variations.order_by("-price", "product")
            selected_sort = {"value": sort, "display": "Price (High to Low)"}
        else:
            selected_sort = {"value": "a-z", "display": "A - Z"}
        selected_sort["url"] = f"sort={sort}"
    else:
        variations = variations.prefetch_related("orderitems").annotate(
            num_orders=Count("orderitems")
        ).order_by("-num_orders", "price")
    
    if not variations:
        return {"products": None, "search_query": q}
    
    paginator = Paginator(variations, 12)
    page_number = request.GET.get("page")
    page_obj = paginator.get_page(page_number)
    return {"products": page_obj, "search_query": q, "search_filter_query": selected_sort}


def get_products(request):
    product = None
    filter_query = '?'
    categories = Category.objects.all()
    if request.GET.get('category'):
        products = Product.objects.filter(categories__slug=request.GET.get('category'))
        filter_query += f"category={request.GET.get('category')}&"
    else:
        products = Product.objects.all()
    if request.GET.get('product'):
        product = Product.objects.filter(slug=request.GET.get('product')).first()
        filter_query += f"product={request.GET.get('product')}&"
    
    if product:
        variations = product.variations.all()
    else:
        variations = Variation.objects.select_related("product").filter(product__in=products)
    variations = variations.select_related("discount").prefetch_related(
        "keyvalues",
        "images",
        "keyvalues__key",
        "keyvalues__value"
    ).filter(availability=True, inventory__gt=0)

    selected_sort = None
    if request.GET.get('sort'):
        sort = request.GET.get('sort')
        if sort == "z-a":
            variations = variations.order_by("-product")
            selected_sort = {"value": sort, "display": "Z - A"}
        elif sort == "price-asc":
            variations = variations.order_by("price", "product")
            selected_sort = {"value": sort, "display": "Price (Low to High)"}
        elif sort == "price-desc":
            variations = variations.order_by("-price", "product")
            selected_sort = {"value": sort, "display": "Price (High to Low)"}
        else:
            selected_sort = {"value": "a-z", "display": "A - Z"}
        filter_query += f"sort={sort}&"
    else:
        variations = variations.prefetch_related("orderitems").annotate(
            num_orders=Count("orderitems")
        ).order_by("-num_orders", "price")

    
    paginator = Paginator(variations, 12)
    page_number = request.GET.get('page')
    page_obj = paginator.get_page(page_number)

    if filter_query == '?':
        filter_query = None

    try:
        selected_category = None if not request.GET.get('category') else Category.objects.get(slug=request.GET.get('category'))
    except:
        selected_category = None
    selected_product = None if not request.GET.get('product') else product

    return {
        "products": page_obj,
        "selected_category": selected_category,
        "selected_product": selected_product,
        "selected_sort": selected_sort,
        "products_set": products,
        "categories": categories,
        "filter_query": filter_query
    }


def validate_email(emailaddress):
    pos_AT = 0
    count_AT = 0
    count_DT = 0
    if emailaddress[0] == '@' or emailaddress[-1] == '@':
        return False
    if emailaddress[0] == '.' or emailaddress[-1] == '.':
        return False
    for c in range(len(emailaddress)):
        if emailaddress[c] == '@':
            pos_AT = c
            count_AT = count_AT + 1
    if count_AT != 1:
        return False
        
    username = emailaddress[0:pos_AT]
    if not username[0].isalnum() or not username[-1].isalnum():
        return False
    for d in range(len(emailaddress)):
        if emailaddress[d] == '.':
            if d == (pos_AT+1):
                return False
            if d > pos_AT:
                word = emailaddress[(pos_AT+1):d]
                if not word.isalnum():
                    return False
                pos_AT = d
                count_DT = count_DT + 1
    if count_DT < 1 or count_DT > 2:
        return False
        
    return True


### is this product available at this pincode
def check_availability(pincode, variation_id, variation_delivery_time_in_days):
    try:
        p = Pincode.objects.get(pincode=pincode)
    except:
        return {"availability": False}
    
    if Pincode.objects.filter(pincode=pincode, products_unavailable__id=variation_id).exists():
        return {"availability": False}
    
    if p.delivery_time_in_days > variation_delivery_time_in_days:
        return {"availability": True, "delivery_time_in_days": p.delivery_time_in_days}
    else:
        return {"availability": True, "delivery_time_in_days": variation_delivery_time_in_days}



def add_item_to_cart(request, variation_id, qty):
    try:
        qty = float(qty)
        variation_id = int(variation_id)
    except:
        return {"success": False}
    
    variation = Variation.objects.select_related("discount", "product").prefetch_related(
        "keyvalues",
        "keyvalues__key",
        "keyvalues__value"
    ).filter(id=variation_id, availability=True, inventory__gt=0).first()

    if variation.inventory < qty:
        return {"success": False, "out_of_stock": f"Available: {variation.inventory} {variation.unit}"}
    
    if not request.session.get("cart", False):
        request.session["cart"] = {
            "cart_items": {},
            "cart_amount": {
                "amount": 0,
                "discount": 0,
                "discount_code": None,
                "amount_payable": 0
            },
            "count": 0,
            "currency": settings.CURRENCY,
            "checkout_url": reverse('store:checkout'),
            "delivery_address": None
        }
    
    vid = str(variation_id)
    zipcode_availability = None
    unit_price = float(variation.price)
    amount = (qty * unit_price)
    amount_payable = amount
    discount = 0
    if variation.discount:
        amount_payable = amount - (amount*(variation.discount.percent/100))
        discount = variation.discount.percent
    
    if vid not in request.session["cart"]["cart_items"]:
        key_value_pair = []
        for kv in variation.keyvalues.all():
            key_value_pair.append({
                "key": kv.key.key.strip().lower().title(),
                "value": kv.value.value.strip().lower().title()
            })
        request.session["cart"]["cart_items"][vid] = {
            "title": variation.product.title.strip().lower().title(),
            "id": variation.id,
            "qty": round(qty, 2),
            "unit_price": round(unit_price, 2),
            "unit": variation.unit,
            "amount": round(amount, 2),
            "discount": discount,
            "amount_payable": round(amount_payable, 2),
            "key_value_pair": key_value_pair,
            "currency": settings.CURRENCY,
            "zipcode_availability": zipcode_availability,
            "delivery_time_in_days": variation.delivery_time_in_days
        }
    else:
        request.session["cart"]["cart_items"][vid]["qty"] = round((request.session["cart"]["cart_items"][vid]["qty"] + qty), 2)
        request.session["cart"]["cart_items"][vid]["amount"] = round((request.session["cart"]["cart_items"][vid]["amount"] + amount), 2)
        request.session["cart"]["cart_items"][vid]["amount_payable"] = round((request.session["cart"]["cart_items"][vid]["amount_payable"] + amount_payable), 2)
        
    if request.session["cart"]["delivery_address"] and settings.CHECK_DELIVERABILITY:
        request.session["cart"]["cart_items"][vid]["zipcode_availability"] = check_availability(
            request.session["cart"]["delivery_address"]["pincode"],
            variation_id,
            request.session["cart"]["cart_items"][vid]["delivery_time_in_days"]
        )

    request.session["cart"]["cart_amount"]["amount"] = round((request.session["cart"]["cart_amount"]["amount"] + amount_payable), 2)
    if request.session["cart"]["cart_amount"]["discount"]:
        d = request.session["cart"]["cart_amount"]["discount"]
        amount_payable = amount_payable - (amount_payable*(d/100))
    request.session["cart"]["cart_amount"]["amount_payable"] = round((request.session["cart"]["cart_amount"]["amount_payable"] + amount_payable), 2)
    request.session["cart"]["count"] = len(request.session["cart"]["cart_items"].keys())
    request.session.modified = True
    return {"success": True, "count": request.session["cart"]["count"]}


def remove_item_from_cart(request, variation_id):
    try:
        variation_id = int(variation_id)
    except ValueError:
        return {"success": False, "message": "Invalid Request"}
    
    if not Variation.objects.filter(id=variation_id).exists():
        return {"success": False, "message": "Invalid Request"}
    
    vid = str(variation_id)
    if vid not in request.session["cart"]["cart_items"]:
        return {"success": False, "message": "Invalid Request"}
    
    item_amount_payable = request.session["cart"]["cart_items"][vid]["amount_payable"]
    cart_amount = request.session["cart"]["cart_amount"]["amount"]
    cart_amount_payable = request.session["cart"]["cart_amount"]["amount_payable"]
    request.session["cart"]["cart_items"].pop(vid)

    request.session["cart"]["cart_amount"]["amount"] = round((cart_amount - item_amount_payable), 2)
    if request.session["cart"]["cart_amount"]["discount"]:
        d = request.session["cart"]["cart_amount"]["discount"]
        item_amount_payable = item_amount_payable - (item_amount_payable*(d/100))
    request.session["cart"]["cart_amount"]["amount_payable"] = round((cart_amount_payable - item_amount_payable), 2)
    request.session["cart"]["count"] = len(request.session["cart"]["cart_items"].keys())
    if request.session["cart"]["count"] == 0:
        request.session["cart"].clear()
        request.session.pop("cart", False)
        request.session.modified = True
        return {"success": True, "cart": False, "count": 0}
    request.session.modified = True
    return {"success": True, "count": request.session["cart"]["count"], "cart": request.session["cart"]}


def get_preferred_delivery_time(pdt):
    # Fri, 23 Sep 2022 00:00:00 GMT
    f = "%a, %d %b %Y %H:%M:%S %Z"
    try:
        dt = datetime.strptime(pdt, f)
    except:
        dt = None
    return dt


def deliveverable_at_location(request, pincode):
    if not Pincode.objects.filter(pincode=pincode).exists():
        return {"availability": False, "deliverability": False}
    
    for item in request.session["cart"]["cart_items"].values():
        if not item["zipcode_availability"]["availability"]:
            return {"availability": False, "deliverability": True}
    
    return {"availability": True, "deliverability": True}


def max_delivery_time(request, pincode):
    p = []
    pin = Pincode.objects.filter(pincode=pincode).first()
    if pin and settings.CHECK_DELIVERABILITY:
        p.append(pin.delivery_time_in_days)
    
    for item in request.session["cart"]["cart_items"].values():
        p.append(item["delivery_time_in_days"])
    return max(p)


def get_all_orders(request):
    orders = Order.objects.select_related("address", "discount", "cancellation").filter(customer=request.user)
    paginator = Paginator(orders, 10)
    page_number = request.GET.get('page')
    page_obj = paginator.get_page(page_number)
    return page_obj


def check_inventory(items):
    for item in items:
        v = Variation.objects.filter(id=item["id"]).first()
        if v and v.inventory < item["qty"]:
            if v.inventory <= 0:
                return {"success": False, "message": f"{item['title']} (ID: {v.id}) is out of stock."}
            return {"success": False, "message": f"Only {v.inventory} {v.unit} of {item['title']} (ID: {v.id}) is available."}
    return {"success": True}



from django.shortcuts import render
from django.http import HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib import messages
from django.db import transaction
from django.utils.timezone import make_aware
from time import sleep
from datetime import datetime, timedelta
from .models import Discount, Order, Product, Address, Pincode, Variation, OrderItem, OrderCancellation
from . import utility
from django.views.decorators.csrf import ensure_csrf_cookie


### Store Homepage
@ensure_csrf_cookie
def homepage(request):
    context = utility.get_products(request)
    return render(
        request,
        "store/homepage.html",
        context
    )


### get category wise products for homepage product filter
def getProducts(request):
    if request.method == "GET":
        return HttpResponseRedirect(reverse('store:homepage'))
    category = request.POST["category"]
    p = []
    if not category:
        return JsonResponse({"products": p})
    
    products = Product.objects.filter(categories__slug=category)
    for product in products:
        p.append({
            "slug": product.slug,
            "title": product.title
        })
    return JsonResponse({"products": p})


### add product to cart
def add2cart(request):
    if request.method == "GET":
        return JsonResponse({"success": False, "message": "Invalid Request"})
    
    qty = request.POST["qty"]
    variation_id = request.POST["variation_id"]

    if not qty or not variation_id:
        return JsonResponse({"success": False})
    
    response = utility.add_item_to_cart(request, variation_id, qty)
    return JsonResponse(response)


### display cart offcanvas
def getCart(request):
    if not request.session.get("cart", False):
        return JsonResponse({"success": False})
    count = len(request.session["cart"]["cart_items"].keys())
    return JsonResponse({"success": True, "count": count, "cart": request.session["cart"]})


def clear_cart_items(request):
    if request.session.get("cart", False):
        request.session["cart"].clear()
        request.session.pop("cart", False)
        request.session.modified = True
    return


### clear cart
def clearCart(request):
    clear_cart_items(request)
    return JsonResponse({"success": True})


### remove product from cart
def removeFromCart(request):
    if request.method == "GET" or not request.session.get("cart", False):
        return JsonResponse({"success": False, "message": "Invalid Request"})
    
    variation_id = request.POST["variation_id"]
    if not variation_id:
        return JsonResponse({"success": False, "message": "Invalid Request"})
    
    response = utility.remove_item_from_cart(request, variation_id)
    return JsonResponse(response)


### apply cart level discount
def apply_discount(request):
    if request.method == "GET" or not request.session.get("cart", False):
        return JsonResponse({"success": False, "message": "Invalid Request"})
    
    code = request.POST["code"]
    if not code:
        return JsonResponse({"success": False, "message": "Invalid Request"})
    
    try:
        discount = Discount.objects.get(code=code)
    except:
        return JsonResponse({"success": False, "message": "Incorrect Code"})
    
    if request.session["cart"]["cart_amount"]["discount"]:
        return JsonResponse({"success": False, "message": "Invalid Request"})
    
    request.session["cart"]["cart_amount"]["discount"] = discount.percent
    request.session["cart"]["cart_amount"]["discount_code"] = discount.code
    amount_payable = request.session["cart"]["cart_amount"]["amount_payable"]
    amount_payable = amount_payable - (amount_payable*(discount.percent/100))
    request.session["cart"]["cart_amount"]["amount_payable"] = round(amount_payable, 2)
    request.session.modified = True
    return JsonResponse({"success": True, "count": request.session["cart"]["count"], "cart": request.session["cart"]})


### checkout page
@ensure_csrf_cookie
@login_required
def checkout(request):
    if not request.session.get("cart", False):
        return HttpResponseRedirect(reverse('store:homepage'))

    return render(
        request,
        "store/checkout.html",
        {"cart": request.session["cart"]}
    )


### select a saved address
@login_required
def selectDeliveryAddress(request):
    if request.method == "GET" or not request.session.get("cart", False):
        return JsonResponse({"success": False})
    
    address_id = request.POST["address_id"]
    if not address_id:
        return JsonResponse({"success": False})
    
    try:
        address_id = int(address_id)
        a = Address.objects.select_related("user").get(id=address_id)
    except:
        return JsonResponse({"success": False})
    
    if a.user != request.user:
        return JsonResponse({"success": False})
    
    address = {
        "id": a.id,
        "editable": a.editable,
        "is_primary": a.is_primary,
        "first_name": a.first_name,
        "last_name": a.last_name,
        "email": a.email,
        "mobile": a.mobile,
        "landline": None if not a.landline else a.landline,
        "address1": a.address1,
        "address2": None if not a.address2 else a.address2,
        "landmark": a.landmark,
        "pincode": a.pincode,
        "city": a.city,
        "state": a.state,
        "country": a.country
    }

    request.session["cart"].pop("delivery_address", None)
    request.session["cart"]["delivery_address"] = address

    if settings.CHECK_DELIVERABILITY:
        items = list(request.session["cart"]["cart_items"].values())
        for item in items:
            vid = str(item["id"])
            request.session["cart"]["cart_items"][vid]["zipcode_availability"] = utility.check_availability(
                a.pincode,
                item["id"],
                request.session["cart"]["cart_items"][vid]["delivery_time_in_days"]
            )
    request.session.modified = True
    return JsonResponse({"success": True, "address": address, "cart": request.session["cart"]})


### get delivery addresses
@login_required
def getDeliveryAddress(request):
    if not request.session.get("cart", False):
        return JsonResponse({"success": False})

    address = request.user.address.all()
    if not address:
        return JsonResponse({"success": True, "address": None})
    add = []
    for a in address:
        if request.session["cart"].get("delivery_address", False) and a.id == request.session["cart"]["delivery_address"]["id"]:
            selected = True
        else:
            selected = False
        add.append({
            "id": a.id,
            "editable": a.editable,
            "is_primary": a.is_primary,
            "first_name": a.first_name,
            "last_name": a.last_name,
            "email": a.email,
            "mobile": a.mobile,
            "landline": None if not a.landline else a.landline,
            "address1": a.address1,
            "address2": None if not a.address2 else a.address2,
            "landmark": a.landmark,
            "pincode": a.pincode,
            "city": a.city,
            "state": a.state,
            "country": a.country,
            "selected_address": selected
        })
    return JsonResponse({"success": True, "address": add})


### remove Selected Delivery Address
@login_required
def removeSelectedDeliveryAddress(request):
    if request.method == "GET" or not request.session.get("cart", False) or not request.session["cart"].get("delivery_address", False):
        return JsonResponse({"success": False})
    
    address_id = request.POST["address_id"]
    if not address_id:
        return JsonResponse({"success": False})
    
    try:
        address_id = int(address_id)
        a = Address.objects.select_related("user").get(id=address_id)
    except:
        return JsonResponse({"success": False})
    
    if a.user != request.user or address_id != request.session["cart"]["delivery_address"]["id"]:
        return JsonResponse({"success": False})
    
    request.session["cart"]["delivery_address"].clear()
    request.session["cart"].pop("delivery_address", None)

    if settings.CHECK_DELIVERABILITY:
        items = list(request.session["cart"]["cart_items"].values())
        for item in items:
            vid = str(item["id"])
            request.session["cart"]["cart_items"][vid]["zipcode_availability"] = None
    request.session.modified = True
    return JsonResponse({"success": True, "cart": request.session["cart"]})


### place order
@login_required
def placeOrder(request):
    if request.method == "GET" or not request.session.get("cart", False) or not request.session["cart"].get("delivery_address", False):
        return JsonResponse({
            "success": False,
            "reload": True,
            "reload_url": reverse("store:homepage")
        })
    
    pdt = request.POST["datetime"]
    payment_method = request.POST["payment_method"]
    if not payment_method:
        return JsonResponse({"success": False, "message": "Please select a payment method."})
    
    if payment_method not in ["CCD", "DCD", "PPL", "OTH", "COD"]:
        clear_cart_items(request)
        return JsonResponse({
            "success": False,
            "reload": True,
            "reload_url": reverse("store:homepage")
        })
    
    dt = None if not pdt else utility.get_preferred_delivery_time(pdt)
    if settings.CHECK_DELIVERABILITY:
        d = utility.deliveverable_at_location(request, request.session["cart"]["delivery_address"]["pincode"])
        if not d["deliverability"]:
            return JsonResponse({
                "success": False,
                "message": "Home delivery is not available at your location."
            })
        elif not d["availability"]:
            return JsonResponse({
                "success": False,
                "message": f'Please remove the items that are not deliverable at your location, {request.session["cart"]["delivery_address"]["pincode"]}, and try again.'
            })
    
    max_time = utility.max_delivery_time(request, request.session["cart"]["delivery_address"]["pincode"])
    if dt and (dt - datetime.now()).days < max_time:
        return JsonResponse({"success": False, "message": f"We cannot deliver your order in less than {max_time} business day(s)."})
    elif not dt:
        dt = datetime.now() + timedelta(days=max_time)
    
    discount_code = None if not request.session["cart"]["cart_amount"]["discount_code"] else request.session["cart"]["cart_amount"]["discount_code"]
    if discount_code:
        try:
            discount = Discount.objects.get(code=discount_code)
        except:
            clear_cart_items(request)
            return JsonResponse({
                "success": False,
                "reload": True,
                "reload_url": reverse("store:homepage")
            })
        else:
            if Order.objects.filter(customer=request.user, discount=discount).exists():
                request.session["cart"]["cart_amount"]["amount_payable"] = request.session["cart"]["cart_amount"]["amount"]
                request.session["cart"]["cart_amount"]["discount_code"] = None
                request.session["cart"]["cart_amount"]["discount"] = 0
                request.session.modified = True
                messages.add_message(request, messages.ERROR, "You cannot use a discount code twice.")
                return JsonResponse({
                    "success": False,
                    "reload": True,
                    "reload_url": reverse("store:checkout")
                })
    else:
        discount = None
    
    address_id = request.session["cart"]["delivery_address"]["id"]
    try:
        address = Address.objects.select_related("user").get(id=address_id)
    except:
        clear_cart_items(request)
        return JsonResponse({
            "success": False,
            "reload": True,
            "reload_url": reverse("store:homepage")
        })
    if address.user != request.user:
        clear_cart_items(request)
        return JsonResponse({
            "success": False,
            "reload": True,
            "reload_url": reverse("store:homepage")
        })
    
    response = utility.check_inventory(list(request.session["cart"]["cart_items"].values()))
    if not response["success"]:
        return JsonResponse(response)
    
    try:
        with transaction.atomic():
            order = Order.objects.create(
                customer = request.user,
                address = address,
                amount = request.session["cart"]["cart_amount"]["amount"],
                discount = discount,
                amount_payable = request.session["cart"]["cart_amount"]["amount_payable"],
                payment_method = payment_method,
                expected_delivery_time = make_aware(dt),
                count = request.session["cart"]["count"]
            )

            for item in request.session["cart"]["cart_items"].values():
                v = Variation.objects.get(id=item["id"])
                order_item = OrderItem.objects.create(
                    order = order,
                    product = v,
                    unit_price = item["unit_price"],
                    quantity = item["qty"],
                    amount = item["amount"],
                    discount_percent = item["discount"],
                    amount_payable = item["amount_payable"]
                )
                v.inventory -= item["qty"]
                if v.inventory <= 0:
                    v.availability = False
                v.save()
            address.editable = False
            address.save()
    except:
        messages.add_message(
            request,
            messages.ERROR,
            "Something went wrong. Please try again later."
        )
        return JsonResponse({
            "success": False,
            "reload": True,
            "reload_url": reverse("store:checkout")
        })
    
    clear_cart_items(request)
    messages.add_message(
        request,
        messages.SUCCESS,
        "Thank you for your order."
    )
    return_url = reverse("store:order", kwargs={'order_id': order.id})
    return JsonResponse({"success": True, "return_url": return_url})


### all orders page
@ensure_csrf_cookie
@login_required
def orders(request):
    orders = utility.get_all_orders(request)
    return render(
        request,
        "store/orders.html",
        {"orders": orders}
    )


### single order page
@ensure_csrf_cookie
@login_required
def order(request, order_id):
    try:
        order_id = int(order_id)
    except ValueError:
        return HttpResponseRedirect(reverse("store:orders"))
    
    order = Order.objects.select_related(
        "address",
        "discount",
        "cancellation",
    ).prefetch_related(
        "orderitems__product__product",
        "orderitems__product__keyvalues__key",
        "orderitems__product__keyvalues__value",
    ).filter(
        id=order_id,
        customer=request.user
    ).first()

    if not order:
        return HttpResponseRedirect(reverse("store:orders"))
    
    return render(
        request,
        "store/order.html",
        {"order": order}
    )


@login_required
def cancelOrder(request):
    if request.method == "GET":
        return JsonResponse({"success": False})
    
    order_id = request.POST["order_id"]
    if not order_id:
        return JsonResponse({"success": False})
    
    try:
        order_id = int(order_id)
    except ValueError:
        return JsonResponse({"success": False})
    
    order = Order.objects.filter(id=order_id, customer=request.user, order_status='O').first()
    if not order:
        return JsonResponse({"success": False})
    
    try:
        with transaction.atomic():
            order.order_status = 'P'
            order.payment_status = 'U'
            order.save()
            
            oc = OrderCancellation.objects.create(
                order = order
            )
    except:
        messages.add_message(
            request,
            messages.ERROR,
            "Something went wrong. Please try again later."
        )
        return JsonResponse({
            "success": True,
            "redirect_url": reverse("store:order", kwargs={'order_id': order_id})
        })
    else:
        messages.add_message(
            request,
            messages.SUCCESS,
            "Your request for cancellation is being processed."
        )
        return JsonResponse({
            "success": True,
            "redirect_url": reverse("store:order", kwargs={'order_id': order_id})
        })


### Search page
@ensure_csrf_cookie
def search(request):
    response = utility.get_search_products(request)
    if not response:
        return HttpResponseRedirect(reverse("store:homepage"))

    return render(
        request,
        "store/search.html",
        {
            "products": response["products"],
            "search_query": response["search_query"],
            "search_filter_query": response["search_filter_query"]
        }
    )


### Products page not configured
@ensure_csrf_cookie
def products(request, product_slug):
    return render(
        request,
        "store/products.html"
    )


### product variations page
@ensure_csrf_cookie
def product(request, product_slug, variation_id):
    try:
        variation_id = int(variation_id)
    except ValueError:
        return HttpResponseRedirect(reverse("store:homepage"))

    variation = Variation.objects.select_related(
        "product",
        "discount"
    ).prefetch_related(
        "product__categories",
        "keyvalues",
        "images",
        "keyvalues__key",
        "keyvalues__value"
    ).filter(
        id = variation_id,
        product__slug = product_slug
    ).first()

    if not variation:
        return HttpResponseRedirect(reverse("store:homepage"))
    return render(
        request,
        "store/product.html",
        {
            "product": variation,
            "product_slug": product_slug,
            "checkDeliverability": settings.CHECK_DELIVERABILITY
        }
    )


def checkDeliverability(request):
    if request.method == "GET":
        return HttpResponseRedirect(reverse("store:homepage"))
    
    pincode  =request.POST["pincode"]
    variation_id = request.POST["variation_id"]

    if not pincode or not variation_id:
        return JsonResponse({"success": False, "message": "Incomplete Form"})
    
    try:
        variation_id = int(variation_id)
    except ValueError:
        return JsonResponse({"success": False, "message": "Invalid Request"})
    
    pincode = Pincode.objects.filter(pincode=pincode).first()
    if not pincode:
        return JsonResponse({"success": False, "message": "Sorry! Home delivery is not available at this location."})
    
    if Pincode.objects.filter(products_unavailable__id=variation_id).exists():
        return JsonResponse({"success": False, "message": "Sorry! This product is not deliverable at this location."})
    

    v = Variation.objects.filter(id=variation_id).first()
    if not v:
        return JsonResponse({"success": False, "message": "Invalid Request"})
    
    m = max([v.delivery_time_in_days, pincode.delivery_time_in_days])
    return JsonResponse({"success": True, "message": f"Available! Delivery Time: {m} business day(s)."})
    

    
    





    



    







    
    
    

    




        

### user profile
@ensure_csrf_cookie
@login_required
def profile(request):
    return render(
        request,
        "store/profile.html"
    )


### get user info
@login_required
def getUser(request):
    user = {
        "authentication_account": reverse("authentication:account"),
        "date_joined": request.user.date_joined,
        "first_name": request.user.first_name,
        "last_name": request.user.last_name,
        "username": request.user.username,
        "email": request.user.email,
        "last_login": request.user.last_login
    }
    address = request.user.address.all()
    if not address:
        return JsonResponse({"success": True, "address": None, "user": user})
    add = []
    for a in address:
        add.append({
            "id": a.id,
            "editable": a.editable,
            "is_primary": a.is_primary,
            "first_name": a.first_name,
            "last_name": a.last_name,
            "email": a.email,
            "mobile": a.mobile,
            "landline": None if not a.landline else a.landline,
            "address1": a.address1,
            "address2": None if not a.address2 else a.address2,
            "landmark": a.landmark,
            "pincode": a.pincode,
            "city": a.city,
            "state": a.state,
            "country": a.country
        })
    return JsonResponse({"success": True, "address": add, "user": user})


### add a delivery address
@login_required
def addAddress(request):
    if request.method == "GET":
        return JsonResponse({"success": False, "message": "Invalid Request"})
    
    first_name = request.POST["first_name"]
    last_name = request.POST["last_name"]
    email = request.POST["email"]
    mobile = request.POST["mobile"]
    landline = request.POST["landline"]
    landmark = request.POST["landmark"]
    address1 = request.POST["address1"]
    address2 = request.POST["address2"]
    city = request.POST["city"]
    pincode = request.POST["pincode"]
    state = request.POST["state"]
    country = request.POST["country"]
    if not first_name or not last_name or not email or not mobile or not landmark or not address1 \
        or not city or not pincode or not state or not country:
        return JsonResponse({"success": False, "message": "Incomplete Form"})
    
    if not utility.validate_email(email):
        return JsonResponse({"success": False, "message": "Invalid Email"})
    
    try:
        a = Address.objects.create(
            user = request.user,
            first_name = first_name.strip().lower().title(),
            last_name = last_name.strip().lower().title(),
            email = email,
            mobile = mobile,
            landline = None if not landline else landline,
            address1 = address1.strip().lower().title(),
            address2 = None if not address2 else address2.strip().lower().title(),
            landmark = landmark.strip().lower().title(),
            pincode = pincode,
            city = city.strip().lower().title(),
            state = state.strip().lower().title(),
            country = country.strip().lower().title()
        )
    except:
        return JsonResponse({"success": False, "message": "Something went wrong. Please try again later."})
    
    address = {
        "id": a.id,
        "editable": a.editable,
        "is_primary": a.is_primary,
        "first_name": a.first_name,
        "last_name": a.last_name,
        "email": a.email,
        "mobile": a.mobile,
        "landline": None if not a.landline else a.landline,
        "address1": a.address1,
        "address2": None if not a.address2 else a.address2,
        "landmark": a.landmark,
        "pincode": a.pincode,
        "city": a.city,
        "state": a.state,
        "country": a.country
    }
    return JsonResponse({"success": True, "address": address})


### set primary address
@login_required
def makeAddressPrimary(request):
    if request.method == "GET":
        return JsonResponse({"success": False})
    
    address_id = request.POST["address_id"]
    if not address_id:
        return JsonResponse({"success": False})
    
    try:
        address_id = int(address_id)
        address = Address.objects.get(id=address_id)
    except:
        return JsonResponse({"success": False})
    
    if address.user != request.user:
        return JsonResponse({"success": False})
    
    addresses = request.user.address.all()
    for a in addresses:
        if a == address:
            a.is_primary = True
        else:
            a.is_primary = False
        a.save()
    
    return JsonResponse({"success": True})


### delete delivery address
@login_required
def deleteAddress(request):
    if request.method == "GET":
        return JsonResponse({"success": False})
    
    address_id = request.POST["address_id"]
    if not address_id:
        return JsonResponse({"success": False})
    
    try:
        address_id = int(address_id)
        address = Address.objects.get(id=address_id)
    except:
        return JsonResponse({"success": False})
    
    if address.user != request.user or not address.editable or address.orders.exists():
        return JsonResponse({"success": False})
        
    try:
        address.delete()
    except:
        return JsonResponse({"success": False})
    else:
        return JsonResponse({"success": True})


### get user delivery address
@login_required
def getAddress(request):
    if request.method == "GET":
        return JsonResponse({"success": False})
    
    address_id = request.POST["address_id"]
    if not address_id:
        return JsonResponse({"success": False})
    
    try:
        address_id = int(address_id)
        a = Address.objects.get(id=address_id)
    except:
        return JsonResponse({"success": False})
    
    if a.user != request.user or not a.editable or a.orders.exists():
        return JsonResponse({"success": False})
    
    address = {
        "id": a.id,
        "first_name": a.first_name,
        "last_name": a.last_name,
        "email": a.email,
        "mobile": a.mobile,
        "landline": None if not a.landline else a.landline,
        "address1": a.address1,
        "address2": None if not a.address2 else a.address2,
        "landmark": a.landmark,
        "pincode": a.pincode,
        "city": a.city,
        "state": a.state,
        "country": a.country
    }
            
    return JsonResponse({"success": True, "address": address})


### edit delivery address
@login_required
def editAddress(request):
    if request.method == "GET":
        return JsonResponse({"success": False, "message":"Invalid Request"})
    
    address_id = request.POST["address_id"]
    if not address_id:
        return JsonResponse({"success": False, "message":"Invalid Request"})
    
    try:
        address_id = int(address_id)
        a = Address.objects.get(id=address_id)
    except:
        return JsonResponse({"success": False, "message":"Invalid Request"})
    
    if a.user != request.user or not a.editable or a.orders.exists():
        return JsonResponse({"success": False, "message":"Invalid Request"})
    
    first_name = request.POST["first_name"]
    last_name = request.POST["last_name"]
    email = request.POST["email"]
    mobile = request.POST["mobile"]
    landline = request.POST["landline"]
    landmark = request.POST["landmark"]
    address1 = request.POST["address1"]
    address2 = request.POST["address2"]
    city = request.POST["city"]
    pincode = request.POST["pincode"]
    state = request.POST["state"]
    country = request.POST["country"]
    if not first_name or not last_name or not email or not mobile or not landmark or not address1 \
        or not city or not pincode or not state or not country:
        return JsonResponse({"success": False, "message": "Incomplete Form"})
    
    if not utility.validate_email(email):
        return JsonResponse({"success": False, "message": "Invalid Email"})
    
    a.first_name = first_name.strip().lower().title()
    a.last_name = last_name.strip().lower().title()
    a.email = email
    a.mobile = mobile
    a.landline = None if not landline else landline
    a.address1 = address1.strip().lower().title()
    a.address2 = None if not address2 else address2.strip().lower().title()
    a.landmark = landmark.strip().lower().title()
    a.pincode = pincode
    a.city = city.strip().lower().title()
    a.state = state.strip().lower().title()
    a.country = country.strip().lower().title()
    try:
        a.save()
    except:
        return JsonResponse({"success": False, "message": "Something went wrong. Please try again later."})
    
    return JsonResponse({"success": True})